/*
 * Verify established TLS connections.
 *
 * Copyright (C) 2011-2013  Simon Ruderich
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "tlsproxy.h"
#include "verify.h"

#include <assert.h>
#include <errno.h>

#include <gnutls/x509.h>


static int get_certificate_path(const char *format,
        const char *hostname, char *buffer, size_t size);


int verify_tls_connection(gnutls_session_t session, const char *hostname) {
    int result;
    char path[TLSPROXY_MAX_PATH_LENGTH];

    size_t size;
    unsigned int status;
    gnutls_x509_crt_t cert;
    const gnutls_datum_t *cert_list;
    unsigned int cert_list_size;
    FILE *file;
    char buffer[66]; /* one line in a PEM file is 64 bytes + '\n' + '\0' */
    char server_cert[16384];
    char stored_cert[16384];

    result = gnutls_certificate_verify_peers2(session, &status);
    /* Verification failed (!= invalid certificate but worse), no need for any
     * more checks. */
    if (result < 0) {
        LOG(WARNING,
            "verify_tls_connection(): gnutls_certificate_verify_peers2() failed: %s",
            gnutls_strerror(result));
        return -1;
    }
    /* Definitely an invalid certificate, abort. */
    if (status & GNUTLS_CERT_EXPIRED
            || status & GNUTLS_CERT_REVOKED
            || status & GNUTLS_CERT_NOT_ACTIVATED
            || status & GNUTLS_CERT_INSECURE_ALGORITHM) {
        LOG(WARNING,
            "verify_tls_connection(): invalid server certificate");
        return -1;
    }

    /* We only handle X509 certificates for now. Let validation fail to
     * prevent an attacker from changing the certificate type to prevent
     * detection. */
    if (gnutls_certificate_type_get(session) != GNUTLS_CRT_X509) {
        LOG(WARNING,
            "verify_tls_connection(): no X509 server certificate");
        return -1;
    }

    /* Get server certificate. */

    result = gnutls_x509_crt_init(&cert);
    if (result < 0) {
        LOG(WARNING,
            "verify_tls_connection(): gnutls_x509_crt_init() failed: %s",
            gnutls_strerror(result));
        return -1;
    }

    cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
    if (cert_list == NULL) {
        LOG(WARNING,
            "verify_tls_connection(): gnutls_certificate_get_peers() failed");
        gnutls_x509_crt_deinit(cert);
        return -1;
    }

    result = gnutls_x509_crt_import(cert, &cert_list[0], GNUTLS_X509_FMT_DER);
    if (result < 0) {
        LOG(WARNING,
            "verify_tls_connection(): gnutls_x509_crt_import() failed: %s",
            gnutls_strerror(result));
        gnutls_x509_crt_deinit(cert);
        return -1;
    }

    /* Export the certificate as PEM to compare it with the known certificate
     * stored on disk. */
    size = sizeof(server_cert);
    result = gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_PEM,
                                    server_cert, &size);
    if (result != GNUTLS_E_SUCCESS) {
        LOG(WARNING,
            "verify_tls_connection(): gnutls_x509_crt_export() failed: %s",
            gnutls_strerror(result));
        gnutls_x509_crt_deinit(cert);
        return -1;
    }
    /* We got the certificate as PEM, free the crt struct. */
    gnutls_x509_crt_deinit(cert);

    /* Open stored server certificate file. */
    if (server_certificate_file(&file, hostname, path, sizeof(path)) != 0) {
        LOG(DEBUG, "server certificate:\n%s", server_cert);
        return -1;
    }

    size = 1; /* '\0' */
    stored_cert[0] = '\0'; /* for strcat() */

    while (fgets(buffer, sizeof(buffer), file) != NULL) {
        size += strlen(buffer);
        /* Make sure the buffer is big enough. */
        if (size >= sizeof(stored_cert)) {
            LOG(WARNING, "verify_tls_connection(): '%s' too big", path);
            fclose(file);

            LOG(DEBUG, "server certificate:\n%s", server_cert);
            return -1;
        }

        strcat(stored_cert, buffer);
    }
    if (ferror(file)) {
        LOG(WARNING,
            "verify_tls_connection(): failed to read from '%s': %s",
            path, strerror(errno));
        fclose(file);

        LOG(DEBUG, "server certificate:\n%s", server_cert);
        return -1;
    }
    fclose(file);

    /* Check if the server certificate matches our stored certificate. */
    if (strcmp(stored_cert, server_cert)) {
        LOG(ERROR,
            "verify_tls_connection(): server certificate changed!",
            path, strerror(errno));

        LOG(WARNING, "server certificate:\n%s", server_cert);
        return -2;
    }

    /* Check that the proxy certificate file exists and is readable for this
     * domain. This ensures we send an "invalid" certificate even if the proxy
     * certificate doesn't exist. */
    if (proxy_certificate_path(hostname, path, sizeof(path)) != 0) {
        return -1;
    }
    file = fopen(path, "r");
    if (file == NULL) {
        LOG(WARNING,
            "verify_tls_connection(): proxy certificate doesn't exist: '%s'",
            path);
        return -1;
    }
    fclose(file);

    return 0;
}


static int get_certificate_path(const char *format,
        const char *hostname, char *buffer, size_t size) {
    int result;

    /* Hostname too long. */
    assert(size > strlen(format));
    if (size - strlen(format) <= strlen(hostname)) {
        LOG(WARNING,
            "get_certificate_path(): hostname too long: '%s'",
            hostname);
        return -1;
    }
    /* Try to prevent path traversals in hostnames. */
    if (strstr(hostname, "..") != NULL) {
        LOG(WARNING,
            "get_certificate_path(): possible path traversal: '%s'",
            hostname);
        return -1;
    }
    /* Safe as format is no user input. */
    result = snprintf(buffer, size, format, hostname);
    if (result < 0) {
        LOG_PERROR(ERROR, "get_certificate_path(): snprintf failed");
        return -1;
    } else if ((size_t)result >= size) {
        LOG(ERROR, "get_certificate_path(): snprintf buffer too short");
        return -1;
    }

    return 0;
}

int proxy_certificate_path(const char *hostname, char *path, size_t size) {
    return get_certificate_path(PROXY_SERVER_CERT_FORMAT,
                                hostname, path, size);
}

int server_certificate_file(FILE **file, const char *hostname,
                            char *path, size_t size) {
    if (get_certificate_path(STORED_SERVER_CERT_FORMAT,
                             hostname, path, size) != 0) {
        LOG_PERROR(ERROR,
                   "server_certificate_file(): failed to get path");
        return -1;
    }

    /* Open the stored certificate file. */
    *file = fopen(path, "rb");
    if (*file == NULL) {
        if (global_passthrough_unknown) {
            LOG(DEBUG,
                "server_certificate_file(): failed to open '%s': %s",
                path, strerror(errno));
        } else {
            LOG(WARNING,
                "server_certificate_file(): failed to open '%s': %s",
                path, strerror(errno));
        }
        /* Couldn't open the file, special case. */
        return -2;
    }

    return 0;
}
